#include <vm/page.h>

/*
 * State for naming pages created using vbi_large_page_alloc().
 */
static kmutex_t		Vboxoff_mutex;
static u_offset_t	Vboxoff;
static vnode_t		Vboxvp;

/*
 * Allocate one large page, of size "pgsz" bytes.  "cansleep" is non-zero
 * if the allocation can block.  Note that even if cansleep is non-zero,
 * the allocation may fail.
 *
 * On success, the physical address corresponding to the page will be
 * stored in *pphys, and the page_t will be returned SE_SHARED-locked.
 * On failure, NULL is returned.
 */
page_t *
vbi_large_page_alloc(uint64_t *pphys, size_t pgsz, uint_t cansleep)
{
	const pgcnt_t	npgs = btop(pgsz);
	const uint_t	pcflags =
	    (cansleep ? PG_WAIT : 0) | PG_NORELOC | PG_EXCL;
	struct seg	kseg;
	page_t		*pplist, *rpp;
	u_offset_t	off;
	uint_t		i;

	ASSERT3U(pgsz, >, PAGESIZE);

	if (page_resv(npgs, KM_NOSLEEP) == 0) {
		return (NULL);
	}

	mutex_enter(&Vboxoff_mutex);
	off = P2ROUNDUP(Vboxoff, pgsz);
	Vboxoff = off + pgsz;
	mutex_exit(&Vboxoff_mutex);

	kseg.s_as = &kas;
	pplist = page_create_va_large(&Vboxvp, off, pgsz,
	    pcflags, &kseg, 0, NULL);
	if (pplist == NULL) {
		page_unresv(npgs);
		return (NULL);
	}

	rpp = pplist;
	for (i = 0; i < npgs; i++) {
		page_t	*pp;

		ASSERT(pplist != NULL);
		pp = pplist;
		page_sub(&pplist, pp);
		page_io_unlock(pp);
	}

	*pphys = ptob(rpp->p_pagenum);
	return (rpp);
}

/*
 * Free one large page that was allocated by vbi_large_page_alloc().
 */
void
vbi_large_page_free(page_t *rpp, size_t pgsz)
{
	const pgcnt_t	npgs = btop(pgsz);
	uint_t		i;

	ASSERT(IS_P2ALIGNED(pp->p_pagenum, npgs));
	ASSERT3U(pgsz, >, PAGESIZE);
	ASSERT3U(page_get_pagesize(pp->p_szc), ==, pgsz);

	for (i = 0; i < npgs; i++) {
		page_t	*const	pp = rpp + i;
		u_offset_t	off = pp->p_offset;

		if (!page_tryupgrade(pp)) {
			page_t	*npp;

			page_unlock(pp);
			npp = page_lookup(&Vboxvp, off, SE_EXCL);
			if (npp != pp) {
				panic("vbi_large_page_free(): lookup of
				    <%p, %lx> returned %p rather than %p",
				    &Vboxvp, off, npp, pp);
			}
		}
	}

	page_destroy_pages(rpp);
	page_unresv(npgs);
}
